/*
 * This file is part of Sounds.
 *
 *   Sounds is free software: you can redistribute it and/or modify
 *   it under the terms of the Lesser GNU General Public License
 *   version 3 (LGPL3) as published by the Free Software Foundation.
 *
 *   Sounds is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the Lesser GNU General Public License
 *   along with Sounds.  If not, see <https://www.gnu.org/licenses/>.
 */

import 'dart:async';

import '../recording_disposition.dart';
import '../sound_recorder.dart';

/// An internal class which manages the RecordingDisposition stream.
/// Its main job is aggreate events coming up from the plugin
/// into a single stream.
class RecordingDispositionManager {
  final SoundRecorder _recorder;
  StreamController<RecordingDisposition> _dispositionController;

  /// tracks the last time we sent an update
  /// We do this as we need to 'merge' the two streams
  /// we recieve from the underlying plugin (duration and db)
  /// into a single stream.
  DateTime lastDispositonUpdate = DateTime.now();
  // The last duration we received from the plugin
  Duration _lastDispositionDuration = Duration(seconds: 0);
  // The last db we recieved from the plugin.
  double _lastDispositionDecibels = 0;

  /// The duration between updates to the stream.
  /// Defaults to [10ms].
  Duration interval = Duration(milliseconds: 10);

  /// ctor
  RecordingDispositionManager(this._recorder);

  /// returns the most recent duration that we have from
  /// the OS.
  Duration get lastDuration => _lastDispositionDuration;

  /// Returns a stream of RecordingDispositions
  /// The stream is a broad cast stream and can be called
  /// multiple times however the [interval] is shared between
  /// all stream.
  /// The [interval] sets the time between stream updates.
  /// This is the minimum [interval] and updates may be less
  /// frequent.
  /// Updates will stop if the recorder is paused.
  Stream<RecordingDisposition> stream({Duration interval}) {
    var subscriptionRequired = false;
    if (_dispositionController == null) {
      _dispositionController = StreamController.broadcast();
      recorderSetDbLevelEnabled(_recorder, enabled: true);
      subscriptionRequired = true;
    }

    /// If the interval has changed then we need to resubscribe.
    if (interval != null && this.interval != interval) {
      subscriptionRequired = true;
    }
    this.interval = interval ?? this.interval;

    // interval has changed or this is the first time througn
    if (subscriptionRequired) {
      recorderSetSubscriptionInterval(_recorder, interval);
      recorderSetDbPeakLevelUpdate(_recorder, interval);
    }
    return _dispositionController.stream;
  }

  /// Internal classes calls this method to notify a change
  /// in the db level.
  void updateDbPeakDisposition(double decibels, {bool force = false}) {
    _lastDispositionDecibels = decibels;

    _trySendDisposition(force: force);
  }

  /// [timePaused] The raw duration from the android/ios subsystem
  /// ignores pauses so we need to subtract any pause time from the
  /// duratin.
  void updateDurationDisposition(Duration duration) {
    _lastDispositionDuration = duration;

    _trySendDisposition();
  }

  /// Sends a disposition if the [interval] has elapsed since
  /// we last sent the data.
  /// Set [force] to force an stream update even if the interval
  /// since the last update hasn't lapsed.
  void _trySendDisposition({bool force = false}) {
    if (_dispositionController != null) {
      if (force ||
          DateTime.now().difference(lastDispositonUpdate).inMilliseconds >=
              interval.inMilliseconds) {
        lastDispositonUpdate = DateTime.now();
        _dispositionController.add(RecordingDisposition(
            _lastDispositionDuration, _lastDispositionDecibels));
      }
    }
  }

  /// Call this method once you have finished with the recording
  /// api so we can release any attached resources.
  void release() {
    if (_dispositionController != null) {
      _dispositionController
        // TODO signal that the stream is closed?
        // ..add(null) // We keep that strange line for backward compatibility
        ..close();
      _dispositionController = null;
    }
  }
}
